@@ -36,7 +36,7 @@ fn fletcher16(data: &[u8]) -> u16 {
 // BoxGame will handle rendering, gamestate, inputs and GGRSRequests
 pub struct BoxGame {
     num_players: usize,
-    game_state: BoxGameState,
+    game_state: BoxState,
     last_checksum: (Frame, u64),
     periodic_checksum: (Frame, u64),
 }
@@ -46,14 +46,14 @@ impl BoxGame {
         assert!(num_players <= 4);
         Self {
             num_players,
-            game_state: BoxGameState::new(num_players),
+            game_state: BoxState::new(num_players),
             last_checksum: (NULL_FRAME, 0),
             periodic_checksum: (NULL_FRAME, 0),
         }
     }
 
     // for each request, call the appropriate function
-    pub fn handle_requests(&mut self, requests: Vec<GGRSRequest>) {
+    pub fn handle_requests(&mut self, requests: Vec<GGRSRequest<BoxState>>) {
         for request in requests {
             match request {
                 GGRSRequest::LoadGameState { cell, .. } => self.load_game_state(cell),
@@ -63,19 +63,22 @@ impl BoxGame {
         }
     }
 
-    // serialize current gamestate, create a checksum
+    // save current gamestate, create a checksum
     // creating a checksum here is only relevant for SyncTestSessions
-    fn save_game_state(&mut self, cell: GameStateCell, frame: Frame) {
+    fn save_game_state(&mut self, cell: GameStateCell<BoxState>, frame: Frame) {
         assert_eq!(self.game_state.frame, frame);
         let buffer = bincode::serialize(&self.game_state).unwrap();
         let checksum = fletcher16(&buffer) as u64;
-        cell.save(GameState::new(frame, Some(buffer), Some(checksum)));
+        cell.save(GameState::new_with_checksum(
+            frame,
+            Some(self.game_state.clone()),
+            checksum,
+        ));
     }
 
-    // deserialize gamestate to load and overwrite current gamestate
-    fn load_game_state(&mut self, cell: GameStateCell) {
-        let state_to_load = cell.load();
-        self.game_state = bincode::deserialize(&state_to_load.buffer.unwrap()).unwrap();
+    // load gamestate and overwrite
+    fn load_game_state(&mut self, cell: GameStateCell<BoxState>) {
+        self.game_state = cell.load().data.expect("No data found.");
     }
 
     fn advance_frame(&mut self, inputs: Vec<GameInput>) {
@@ -96,7 +99,7 @@ impl BoxGame {
     pub fn render(&self) {
         clear_background(BLACK);
 
-        // render players TODO: rotate players?
+        // render players
         for i in 0..self.num_players {
             let color = match i {
                 0 => GOLD,
@@ -140,7 +143,6 @@ impl BoxGame {
     pub fn local_input(&self, handle: PlayerHandle) -> Vec<u8> {
         let mut input: u8 = 0;
 
-        // ugly, but it works...
         // player 1 with WASD
         if handle == 0 {
             if is_key_down(KeyCode::W) {
@@ -182,8 +184,8 @@ impl BoxGame {
 }
 
 // BoxGameState holds all relevant information about the game state
-#[derive(Serialize, Deserialize)]
-struct BoxGameState {
+#[derive(Clone, Serialize, Deserialize)]
+pub struct BoxState {
     pub frame: i32,
     pub num_players: usize,
     pub positions: Vec<(f32, f32)>,
@@ -191,7 +193,7 @@ struct BoxGameState {
     pub rotations: Vec<f32>,
 }
 
-impl BoxGameState {
+impl BoxState {
     pub fn new(num_players: usize) -> Self {
         let mut positions = Vec::new();
         let mut velocities = Vec::new();
@@ -270,7 +272,7 @@ impl BoxGameState {
             let mut x = old_x + vel_x;
             let mut y = old_y + vel_y;
 
-            // constrain boxes to canvas borders
+            // constrain players to canvas borders
             x = x.max(0.0);
             x = x.min(WINDOW_WIDTH);
             y = y.max(0.0);
