@@ -6,7 +6,7 @@ use structopt::StructOpt;
 
 const FPS: f64 = 60.0;
 const INPUT_SIZE: usize = std::mem::size_of::<u8>();
-const MAX_PRED_FRAME: usize = 12;
+const MAX_PRED_FRAME: usize = 8;
 
 mod box_game;
 
@@ -117,25 +117,11 @@ async fn main() -> Result<(), Box<dyn std::error::Error>> {
                 accumulator = accumulator.saturating_sub(Duration::from_secs_f64(fps_delta));
 
                 match sess.advance_frame(local_handle, &game.local_input(0)) {
-                    Ok(requests) => {
-                        game.handle_requests(requests);
-                        if sess.current_frame() % (2 * FPS as i32) == 0 {
-                            for p in 0..num_players {
-                                if p == local_handle {
-                                    continue;
-                                }
-                                println!(
-                                    "Frame {}: Player {}, Ping {}",
-                                    sess.current_frame(),
-                                    p,
-                                    sess.network_stats(p).unwrap().ping
-                                );
-                            }
-                        }
-                    }
+                    Ok(requests) => game.handle_requests(requests),
                     Err(GGRSError::PredictionThreshold) => {
                         println!("Frame {} skipped", sess.current_frame())
                     }
+
                     Err(e) => return Err(Box::new(e)),
                 }
             }
