@@ -26,8 +26,8 @@ impl UdpNonBlockingSocket {
     }
 }
 
-impl NonBlockingSocket for UdpNonBlockingSocket {
-    fn send_to(&mut self, msg: &UdpMessage, addr: SocketAddr) {
+impl NonBlockingSocket<SocketAddr> for UdpNonBlockingSocket {
+    fn send_to(&mut self, msg: &UdpMessage, addr: &SocketAddr) {
         let buf = bincode::serialize(&msg).unwrap();
         self.socket.send_to(&buf, addr).unwrap();
     }
