@@ -88,8 +88,10 @@ pub(crate) enum MessageBody {
     KeepAlive,
 }
 
+/// A messages that `NonBlockingSocket` sends and receives. When implementing `NonBlockingSocket`,
+/// you should deserialize received messages into this type and pass them.
 #[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
-pub(crate) struct UdpMessage {
-    pub header: MessageHeader,
-    pub body: MessageBody,
+pub struct UdpMessage {
+    pub(crate) header: MessageHeader,
+    pub(crate) body: MessageBody,
 }
