@@ -13,7 +13,7 @@ use crate::{Frame, PlayerHandle, NULL_FRAME};
 
 use instant::{Duration, Instant};
 use std::collections::vec_deque::Drain;
-use std::collections::VecDeque;
+use std::collections::{HashMap, VecDeque};
 use std::convert::TryFrom;
 use std::net::SocketAddr;
 use std::ops::Add;
@@ -54,7 +54,7 @@ enum ProtocolState {
 }
 
 #[derive(Debug)]
-pub(crate) struct UdpProtocol {
+pub(crate) struct UdpProtocol<A: Eq = SocketAddr> {
     handle: PlayerHandle,
     magic: u16,
     send_queue: VecDeque<UdpMessage>,
@@ -76,14 +76,15 @@ pub(crate) struct UdpProtocol {
     fps: u32,
 
     // the other client
-    peer_addr: SocketAddr,
+    peer_addr: A,
     remote_magic: u16,
     peer_connect_status: Vec<ConnectionStatus>,
 
     // input compression
     pending_output: VecDeque<GameInput>,
-    last_received_input: GameInput,
     last_acked_input: GameInput,
+    max_prediction: usize,
+    recv_inputs: HashMap<Frame, GameInput>,
 
     // time sync
     time_sync_layer: TimeSync,
@@ -99,19 +100,20 @@ pub(crate) struct UdpProtocol {
     last_recv_time: Instant,
 }
 
-impl PartialEq for UdpProtocol {
+impl<A: Eq> PartialEq for UdpProtocol<A> {
     fn eq(&self, other: &Self) -> bool {
         self.handle == other.handle
     }
 }
-impl Eq for UdpProtocol {}
+impl<A: Eq> Eq for UdpProtocol<A> {}
 
-impl UdpProtocol {
+impl<A: Eq> UdpProtocol<A> {
     pub(crate) fn new(
         handle: PlayerHandle,
-        peer_addr: SocketAddr,
+        peer_addr: A,
         num_players: u32,
         input_size: usize,
+        max_prediction: usize,
     ) -> Self {
         let mut magic = rand::random::<u16>();
         while magic == 0 {
@@ -124,6 +126,10 @@ impl UdpProtocol {
             peer_connect_status.push(ConnectionStatus::default());
         }
 
+        // received input history
+        let mut recv_inputs = HashMap::new();
+        recv_inputs.insert(NULL_FRAME, GameInput::blank_input(input_size));
+
         Self {
             handle,
             magic,
@@ -152,8 +158,9 @@ impl UdpProtocol {
 
             // input compression
             pending_output: VecDeque::with_capacity(PENDING_OUTPUT_SIZE),
-            last_received_input: GameInput::blank_input(input_size),
             last_acked_input: GameInput::blank_input(input_size),
+            max_prediction,
+            recv_inputs,
 
             // time sync
             time_sync_layer: TimeSync::new(),
@@ -170,20 +177,17 @@ impl UdpProtocol {
         }
     }
 
-    pub(crate) const fn player_handle(&self) -> PlayerHandle {
+    pub(crate) fn player_handle(&self) -> PlayerHandle {
         self.handle
     }
 
     pub(crate) fn update_local_frame_advantage(&mut self, local_frame: Frame) {
-        if local_frame == NULL_FRAME {
-            return;
-        }
-        if self.last_received_input.frame == NULL_FRAME {
+        if local_frame == NULL_FRAME || self.last_recv_frame() == NULL_FRAME {
             return;
         }
         // Estimate which frame the other client is on by looking at the last frame they gave us plus some delta for the packet roundtrip time.
         let ping = i32::try_from(self.round_trip_time / 2).expect("Ping is higher than i32::MAX");
-        let remote_frame = self.last_received_input.frame + ((ping * self.fps as i32) / 1000);
+        let remote_frame = self.last_recv_frame() + ((ping * self.fps as i32) / 1000);
         // Our frame "advantage" is how many frames behind the remote client we are. (It's an advantage because they will have to predict more often)
         self.local_frame_advantage = remote_frame - local_frame;
     }
@@ -235,7 +239,7 @@ impl UdpProtocol {
         self.state == ProtocolState::Running
     }
 
-    pub(crate) fn is_handling_message(&self, addr: &SocketAddr) -> bool {
+    pub(crate) fn is_handling_message(&self, addr: &A) -> bool {
         self.peer_addr == *addr
     }
 
@@ -335,14 +339,14 @@ impl UdpProtocol {
      *  SENDING MESSAGES
      */
 
-    pub(crate) fn send_all_messages(&mut self, socket: &mut Box<dyn NonBlockingSocket>) {
+    pub(crate) fn send_all_messages(&mut self, socket: &mut Box<dyn NonBlockingSocket<A>>) {
         if self.state == ProtocolState::Shutdown {
             self.send_queue.drain(..);
             return;
         }
 
         for msg in self.send_queue.drain(..) {
-            socket.send_to(&msg, self.peer_addr);
+            socket.send_to(&msg, &self.peer_addr);
         }
     }
 
@@ -380,26 +384,24 @@ impl UdpProtocol {
                     || self.last_acked_input.frame + 1 == input.frame
             );
             body.start_frame = input.frame;
-        } else {
-            body.start_frame = 0;
+
+            // encode all pending inputs to a byte buffer
+            body.bytes = encode(&self.last_acked_input, self.pending_output.iter());
+
+            // the byte buffer should not exceed a certain size to guarantee a maximum UDP packet size
+            assert!(body.bytes.len() <= MAX_PAYLOAD);
+
+            body.ack_frame = self.last_recv_frame();
+            body.disconnect_requested = self.state == ProtocolState::Disconnected;
+            body.peer_connect_status = connect_status.to_owned();
+
+            self.queue_message(MessageBody::Input(body));
         }
-
-        // encode all pending inputs to a byte buffer
-        body.bytes = encode(&self.last_acked_input, self.pending_output.iter());
-
-        // the byte buffer should not exceed a certain size to guarantee a maximum UDP packet size
-        assert!(body.bytes.len() <= MAX_PAYLOAD);
-
-        body.ack_frame = self.last_received_input.frame;
-        body.disconnect_requested = self.state == ProtocolState::Disconnected;
-        body.peer_connect_status = connect_status.to_owned();
-
-        self.queue_message(MessageBody::Input(body));
     }
 
     fn send_input_ack(&mut self) {
         let body = InputAck {
-            ack_frame: self.last_received_input.frame,
+            ack_frame: self.last_recv_frame(),
         };
 
         self.queue_message(MessageBody::InputAck(body));
@@ -539,37 +541,45 @@ impl UdpProtocol {
             }
         }
 
-        // this input has not been encoded with what we expect, so we drop the whole thing
-        // TODO: this could be made so much more efficient if we kept more received input history
-        // so we can properly decode with the right reference
+        // if the encoded packet is decoded with an input we did not receive yet, we cannot recover
         assert!(
-            self.last_received_input.frame == NULL_FRAME
-                || self.last_received_input.frame + 1 >= body.start_frame
+            self.last_recv_frame() == NULL_FRAME || self.last_recv_frame() + 1 >= body.start_frame
         );
-        if self.last_received_input.frame != NULL_FRAME
-            && self.last_received_input.frame + 1 != body.start_frame
-        {
-            return;
-        }
 
-        self.running_last_input_recv = Instant::now();
+        // if we did not receive any input yet, we decode with the blank input,
+        // otherwise we use the input previous to the start of the encoded inputs
+        let decode_frame = if self.last_recv_frame() == NULL_FRAME {
+            NULL_FRAME
+        } else {
+            body.start_frame - 1
+        };
 
-        // we know everything is correct, so we decode
-        let recv_inputs = decode(&self.last_received_input, body.start_frame, &body.bytes)
-            .expect("decoding failed");
+        // if we have the necessary input saved, we decode
+        if let Some(decode_inp) = self.recv_inputs.get(&decode_frame) {
+            self.running_last_input_recv = Instant::now();
 
-        for game_input in &recv_inputs {
-            // skip inputs that we don't need
-            if game_input.frame <= self.last_received_input.frame {
-                continue;
+            let recv_inputs =
+                decode(decode_inp, body.start_frame, &body.bytes).expect("decoding failed");
+
+            for game_input in &recv_inputs {
+                // skip inputs that we don't need
+                if game_input.frame <= self.last_recv_frame() {
+                    continue;
+                }
+                // send the input to the session
+                self.recv_inputs
+                    .insert(game_input.frame, game_input.clone());
+                self.event_queue.push_back(Event::Input(game_input.clone()));
             }
-            // send the input to the session
-            self.last_received_input = game_input.clone();
-            self.event_queue.push_back(Event::Input(game_input.clone()));
-        }
 
-        // send an input ack
-        self.send_input_ack();
+            // send an input ack
+            self.send_input_ack();
+
+            // delete reveiced inputs that are too old
+            let last_recv_frame = self.last_recv_frame();
+            self.recv_inputs
+                .retain(|&k, _| k >= last_recv_frame - 2 * self.max_prediction as i32);
+        }
     }
 
     /// Upon receiving a `InputAck`, discard the oldest buffered input including the acked input.
@@ -590,4 +600,12 @@ impl UdpProtocol {
         assert!(millis >= body.pong);
         self.round_trip_time = millis - body.pong;
     }
+
+    /// Returns the frame of the last received input
+    fn last_recv_frame(&self) -> Frame {
+        match self.recv_inputs.iter().max_by_key(|&(k, _)| k) {
+            Some((k, _)) => *k,
+            None => NULL_FRAME,
+        }
+    }
 }
