@@ -24,15 +24,15 @@ pub(crate) const DEFAULT_DISCONNECT_NOTIFY_START: Duration = Duration::from_mill
 pub(crate) const DEFAULT_FPS: u32 = 60;
 
 #[derive(Debug, PartialEq, Eq)]
-enum Player {
+enum Player<A: Eq = SocketAddr> {
     Local,
-    Remote(Box<UdpProtocol>),
-    Spectator(Box<UdpProtocol>),
+    Remote(Box<UdpProtocol<A>>),
+    Spectator(Box<UdpProtocol<A>>),
 }
 
-impl Player {
+impl<A: Eq> Player<A> {
     #[allow(dead_code)]
-    const fn as_endpoint(&self) -> Option<&UdpProtocol> {
+    fn as_endpoint(&self) -> Option<&UdpProtocol<A>> {
         match self {
             Player::Remote(endpoint) => Some(endpoint),
             Player::Spectator(endpoint) => Some(endpoint),
@@ -40,7 +40,7 @@ impl Player {
         }
     }
 
-    fn as_endpoint_mut(&mut self) -> Option<&mut UdpProtocol> {
+    fn as_endpoint_mut(&mut self) -> Option<&mut UdpProtocol<A>> {
         match self {
             Player::Remote(endpoint) => Some(endpoint),
             Player::Spectator(endpoint) => Some(endpoint),
@@ -48,28 +48,28 @@ impl Player {
         }
     }
 
-    const fn remote_as_endpoint(&self) -> Option<&UdpProtocol> {
+    fn remote_as_endpoint(&self) -> Option<&UdpProtocol<A>> {
         match self {
             Player::Remote(endpoint) => Some(endpoint),
             Player::Spectator(_) | Player::Local => None,
         }
     }
 
-    fn remote_as_endpoint_mut(&mut self) -> Option<&mut UdpProtocol> {
+    fn remote_as_endpoint_mut(&mut self) -> Option<&mut UdpProtocol<A>> {
         match self {
             Player::Remote(endpoint) => Some(endpoint),
             Player::Spectator(_) | Player::Local => None,
         }
     }
 
-    const fn spectator_as_endpoint(&self) -> Option<&UdpProtocol> {
+    fn spectator_as_endpoint(&self) -> Option<&UdpProtocol<A>> {
         match self {
             Player::Spectator(endpoint) => Some(endpoint),
             Player::Remote(_) | Player::Local => None,
         }
     }
 
-    fn spectator_as_endpoint_mut(&mut self) -> Option<&mut UdpProtocol> {
+    fn spectator_as_endpoint_mut(&mut self) -> Option<&mut UdpProtocol<A>> {
         match self {
             Player::Spectator(endpoint) => Some(endpoint),
             Player::Remote(_) | Player::Local => None,
@@ -94,8 +94,7 @@ pub(crate) enum Event {
 }
 
 /// A `P2PSession` provides a UDP protocol to connect to remote clients in a peer-to-peer fashion.
-#[derive(Debug)]
-pub struct P2PSession {
+pub struct P2PSession<T: Clone = Vec<u8>, A: Eq = SocketAddr> {
     /// The number of players of the session.
     num_players: u32,
     /// The number of bytes an input uses.
@@ -103,7 +102,7 @@ pub struct P2PSession {
     /// The maximum number of frames GGRS will roll back. Every gamestate older than this is guaranteed to be correct.
     max_prediction: usize,
     /// The sync layer handles player input queues and provides predictions.
-    sync_layer: SyncLayer,
+    sync_layer: SyncLayer<T>,
     /// FPS defines the expected update frequency of this session.
     fps: u32,
     /// With sparse saving, the session will only request to save the minimum confirmed frame.
@@ -120,9 +119,9 @@ pub struct P2PSession {
     state: SessionState,
 
     /// The `P2PSession` uses this socket to send and receive all messages for remote players.
-    socket: Box<dyn NonBlockingSocket>,
+    socket: Box<dyn NonBlockingSocket<A>>,
     /// A map of player handle to a player struct that handles receiving and sending messages for remote players, remote spectators and register local players.
-    players: HashMap<PlayerHandle, Player>,
+    players: HashMap<PlayerHandle, Player<A>>,
     /// This struct contains information about remote players, like connection status and the frame of last received input.
     local_connect_status: Vec<ConnectionStatus>,
 
@@ -137,7 +136,7 @@ pub struct P2PSession {
     event_queue: VecDeque<GGRSEvent>,
 }
 
-impl P2PSession {
+impl<T: Clone> P2PSession<T> {
     /// Creates a new `P2PSession` for players who participate on the game input. After creating the session, add local and remote players,
     /// set input delay for local players and then start the session.
     /// # Example
@@ -149,7 +148,7 @@ impl P2PSession {
     /// let num_players : u32 = 2;
     /// let max_pred : usize = 8;
     /// let input_size : usize = std::mem::size_of::<u32>();
-    /// let mut session = P2PSession::new(num_players, input_size, max_pred, local_port)?;
+    /// let mut session = P2PSession::<Vec<u8>>::new(num_players, input_size, max_pred, local_port)?;
     /// # Ok(())
     /// # }
     /// ```
@@ -175,14 +174,16 @@ impl P2PSession {
             socket,
         ))
     }
+}
 
+impl<T: Clone, A: Eq> P2PSession<T, A> {
     /// Creates a new `P2PSession` for players who participate on the game input. After creating the session, add local and remote players,
     /// set input delay for local players and then start the session. The session will use the provided socket.
     pub fn new_with_socket(
         num_players: u32,
         input_size: usize,
         max_prediction: usize,
-        socket: impl NonBlockingSocket + 'static,
+        socket: impl NonBlockingSocket<A> + 'static,
     ) -> Self {
         Self::new_impl(num_players, input_size, max_prediction, Box::new(socket))
     }
@@ -191,7 +192,7 @@ impl P2PSession {
         num_players: u32,
         input_size: usize,
         max_prediction: usize,
-        socket: Box<dyn NonBlockingSocket>,
+        socket: Box<dyn NonBlockingSocket<A>>,
     ) -> Self {
         // local connection status
         let mut local_connect_status = Vec::new();
@@ -231,7 +232,7 @@ impl P2PSession {
     /// - Returns `InvalidRequest` when adding more than one local player
     pub fn add_player(
         &mut self,
-        player_type: PlayerType,
+        player_type: PlayerType<A>,
         player_handle: PlayerHandle,
     ) -> Result<PlayerHandle, GGRSError> {
         // currently, you can only add players in the init phase
@@ -333,7 +334,7 @@ impl P2PSession {
         &mut self,
         local_player_handle: PlayerHandle,
         local_input: &[u8],
-    ) -> Result<Vec<GGRSRequest>, GGRSError> {
+    ) -> Result<Vec<GGRSRequest<T>>, GGRSError> {
         // receive info from remote players, trigger events and send messages
         self.poll_remote_clients();
 
@@ -392,9 +393,10 @@ impl P2PSession {
             }
 
             // after all this, we should have saved the confirmed state
-            assert_eq!(
-                self.sync_layer.last_saved_frame(),
-                std::cmp::min(confirmed_frame, self.sync_layer.current_frame())
+            assert!(
+                confirmed_frame == NULL_FRAME
+                    || self.sync_layer.last_saved_frame()
+                        == std::cmp::min(confirmed_frame, self.sync_layer.current_frame())
             );
         }
 
@@ -659,12 +661,12 @@ impl P2PSession {
     }
 
     /// Returns the current frame of a session.
-    pub const fn current_frame(&self) -> Frame {
+    pub fn current_frame(&self) -> Frame {
         self.sync_layer.current_frame()
     }
 
     /// Returns the maximum prediction window of a session.
-    pub const fn max_prediction(&self) -> usize {
+    pub fn max_prediction(&self) -> usize {
         self.max_prediction
     }
 
@@ -677,7 +679,7 @@ impl P2PSession {
     }
 
     /// Returns the current `SessionState` of a session.
-    pub const fn current_state(&self) -> SessionState {
+    pub fn current_state(&self) -> SessionState {
         self.state
     }
 
@@ -687,17 +689,17 @@ impl P2PSession {
     }
 
     /// Returns the number of players this session was constructed with.
-    pub const fn num_players(&self) -> u32 {
+    pub fn num_players(&self) -> u32 {
         self.num_players
     }
 
     /// Returns the input size this session was constructed with.
-    pub const fn input_size(&self) -> usize {
+    pub fn input_size(&self) -> usize {
         self.input_size
     }
 
     /// Returns the number of frames this session is estimated to be ahead of other sessions
-    pub const fn frames_ahead(&self) -> i32 {
+    pub fn frames_ahead(&self) -> i32 {
         self.frames_ahead
     }
 
@@ -727,7 +729,7 @@ impl P2PSession {
     fn add_remote_player(
         &mut self,
         player_handle: PlayerHandle,
-        addr: SocketAddr,
+        addr: A,
     ) -> Result<PlayerHandle, GGRSError> {
         // check if valid player
         if player_handle >= self.num_players as PlayerHandle {
@@ -742,7 +744,13 @@ impl P2PSession {
         }
 
         // create a udp protocol endpoint that handles all the messaging to that remote player
-        let mut endpoint = UdpProtocol::new(player_handle, addr, self.num_players, self.input_size);
+        let mut endpoint = UdpProtocol::new(
+            player_handle,
+            addr,
+            self.num_players,
+            self.input_size,
+            self.max_prediction,
+        );
         endpoint.set_disconnect_notify_start(self.disconnect_notify_start);
         endpoint.set_disconnect_timeout(self.disconnect_timeout);
 
@@ -758,7 +766,7 @@ impl P2PSession {
     fn add_spectator(
         &mut self,
         player_handle: PlayerHandle,
-        addr: SocketAddr,
+        addr: A,
     ) -> Result<PlayerHandle, GGRSError> {
         let spectator_handle = player_handle + 1000;
 
@@ -775,6 +783,7 @@ impl P2PSession {
             addr,
             self.num_players,
             self.input_size * self.num_players as usize,
+            self.max_prediction,
         );
         endpoint.set_disconnect_notify_start(self.disconnect_notify_start);
         endpoint.set_disconnect_timeout(self.disconnect_timeout);
@@ -840,7 +849,7 @@ impl P2PSession {
         &mut self,
         first_incorrect: Frame,
         min_confirmed: Frame,
-        requests: &mut Vec<GGRSRequest>,
+        requests: &mut Vec<GGRSRequest<T>>,
     ) {
         let current_frame = self.sync_layer.current_frame();
         // determine the frame to load
